#ifndef CX__H
#define CX__H

class CX // Using class RE defined in <re.h>
{
  RE x, y;
public:
  const RE& Re () const { return x; }
  const RE& Im () const { return x; }
  // Construction and conversion operations
  ~CX() { }
  CX () { } // Constructeur par dfaut
  CX (RE a)                            : x(a), y(0) { }
  CX (RE a,          RE b)             : x(a), y(b) { }
  CX (int a,         int b = 0)        : x(a), y(b) { }
  CX (long a,        long b = 0)       : x(a), y(b) { }
  CX (float a,       float b = 0)      : x(a), y(b) { }
  CX (double a,      double b = 0)     : x(a), y(b) { }
  CX (const char *a, const char *b)    : x(a), y(b) { }
  CX (const char *buffer);
  operator char* (void) const;
  friend void ctodbl(double& cx, double& cy, const CX& c);
  // Fundamental comparison operators
  friend bool operator == (const CX& a, const CX& b);
  friend bool IsZero (const CX& a);
  friend bool IsOne  (const CX& a);
  friend bool IsReal (const CX& a);
  friend bool NearZero (const CX& a); // proximit de 0
  // Fundamental arithmetic operators
  friend CX operator - (const CX& a);
  friend CX   inv (const CX& a); // returns the inverse
  friend CX  conj (const CX& a); // returns the conjugate
  friend RE   dot (const CX& a, const CX& b); // returns the dot product
  friend RE  norm (const CX& a) { return dot (a, a); }
  friend double arg (const CX& a);
  CX& operator += (const CX& a);
  CX& operator *= (const CX& a);
  CX& operator -= (const CX& a) { *this += -a; return *this; }
  CX& operator /= (const CX& a) { *this *= inv(a); return *this; }
#ifdef mplib_pari
  // Specific for Pari
  GEN genvalue () const {
    GEN g = cgetg(3,t_COMPLEX);
    ((GEN*)g)[1] = x.genvalue();
    ((GEN*)g)[2] = y.genvalue();
    return g;
  }
  friend void genptr (GEN* gptr[], ulong i, CX& c) {
    gptr[2*i  ] = c.x.genptr();
    gptr[2*i+1] = c.y.genptr();
  }
#endif
};

// Construction and conversion operations
// [cx-io.c] CX::CX (const char *buffer);
// [cx-io.c] CX::operator char* (void) const;
// Fundamental comparison operators
inline bool IsZero (const CX& a) { return IsZero(a.x) && IsZero(a.y); }
inline bool IsOne  (const CX& a) { return IsOne (a.x) && IsZero(a.y); }
inline bool IsReal (const CX& a) { return IsZero(a.y); }
inline bool NearZero (const CX& a) { return NearZero(a.x) && NearZero(a.y); }
inline bool operator == (const CX& a, const CX& b)
  { return (a.x == b.x) && (a.y == b.y); }
// Fundamental arithmetic operators
inline CX operator - (const CX& a) { return CX( - a.x, - a.y ); }
inline CX conj (const CX& a) { return CX( a.x, -a.y ); }
inline RE dot (const CX& a, const CX& b) { return a.x * b.x + a.y * b.y; }
inline double arg (const CX& a)
{
  if (IsZero(a)) return NaN;
  if (IsZero(a.x))
    if (a.y > RE(0)) return 0.;
    else             return .5;
  if (IsZero(a.y))
    if (a.x > RE(0)) return .25;
    else             return .75;
  double DeuxPi = 3.1415927 * 2;
  double tangente = a.y / a.x;
  if (a.x > RE(0))
    return atan (tangente) / DeuxPi;
  else
    return atan (tangente) / DeuxPi + .5;
}
inline CX inv (const CX& a)
{
  RE norme = norm (a);
  if (IsZero (norme)) { Error("CX: division by zero"); throw norme; }
  return CX( a.x / norme, - a.y / norme );
}
inline CX& CX::operator += (const CX& a) { x += a.x; y += a.y; return *this; }
inline CX& CX::operator *= (const CX& a)
{
  const RE r = x;
  x *= a.x; x -= a.y * y;
  y *= a.x; y += a.y * r;
  return *this;
}


// -------- Definitions from fundamental methods --------
// Standard arithmetic operators
inline CX operator + (const CX& a, const CX& b) {CX c = a;c += b; return c;}
inline CX operator - (const CX& a, const CX& b) {CX c = a;c -= b; return c;}
inline CX operator * (const CX& a, const CX& b) {CX c = a;c *= b; return c;}
inline CX operator / (const CX& a, const CX& b) {CX c = a;c /= b; return c;}
// Other operators
inline CX sqr (const CX& a) { return a * a; }
inline bool operator != (const CX& a, const CX& b) { return !(a == b); }
inline CX unit (const CX& a)
  { if (!IsZero(a)) return a/CX(sqrt(norm(a))); else return a; }

inline CX pow (const CX& c, const unsigned long exposant)
{
  CX res = 1;
  unsigned long k = exposant;
  CX a = c;
  for ( ; k > 1; k >>= 1) { if (k&1) res *= a; a *= a; }
  if (k) res *= a;
  return res;
}

inline CX direction (const double d)
{
  double DeuxPi = 3.1415927 * 2;
  return CX ( cos(DeuxPi*d), sin(DeuxPi*d) );
}
inline CX root (const CX& a, const unsigned long exposant)
{
  return root (norm (a), 2*exposant) * direction (arg (a) / exposant);
}

#endif
