#include "dessin.h"
#include "solve.h"

// ----------------------------------------------------------------------
// class FILEREAD

class FILEREAD : public ifstream
{
// content of the class
  const char* filename;
  ulong linecount;
  ulong buflen;
  char* buffer;
// member functions and friends
  public:
  FILEREAD(const char* name, int mode=ios::in, int prot=0664)
  : ifstream(name, mode, prot), filename(name), linecount(0),
    buflen(16384), buffer(new char[buflen])
    { if (!(*this)) { cerr << "Erreur pour la lecture du fichier " << filename;
      perror(" "); } }
  ~FILEREAD() { close(); delete[] buffer; }
  void close() { ifstream::close(); buffer[0] = '\0'; }
  int operator!() { return ifstream::operator!(); }
  const char* scanline();
};

const char*
FILEREAD::scanline()
{
  if (!*this) return buffer;
  static int ret = 0;
  char eol;
  while (ret<1) {
    linecount++;
    get( buffer, buflen, '\n');
    if (!get(eol)) { // EOF
      close();
      return buffer;
    }
    else if (eol == '\n') {
      ret = 2;
    }
    else {
      ret = 2;
      ulong linelen = buflen;
      while (get(eol) && (eol!='\n')) linelen++;
      if (eol=='\n') linelen++;
      cerr << "La ligne de donnes " << linecount
           << " est trop longue (" << linelen << " > " << buflen << ")\n";
    }
    if (buffer[0] == '#') ret = 0;
  }
  ret = 0;
  return buffer;
}

// ----------------------------------------------------------------------
// class TOPOLOGIE

TOPOLOGIE::~TOPOLOGIE ()
{
  (*refcount) --;
  if (!*refcount) {
    delete topo_faces;
    delete topo_sommets;
    delete topo_aretes;
    delete topo_quatriemes;
    delete val_quatrieme;
    delete refcount;
  }
}

TOPOLOGIE::TOPOLOGIE (const TOPOLOGIE& t)
{
  topo_genre      = t.topo_genre;
  topo_degre      = t.topo_degre;
  topo_faces      = t.topo_faces;
  topo_sommets    = t.topo_sommets;
  topo_aretes     = t.topo_aretes;
  topo_quatriemes = t.topo_quatriemes;
  val_quatrieme   = t.val_quatrieme;
  refcount        = t.refcount;
  (*refcount) ++;
}

void
TOPOLOGIE::copy_if_multiref()
{
  if (*refcount == 1) return;
  topo_faces      = new VALENCES(*topo_faces);
  topo_sommets    = new VALENCES(*topo_sommets);
  topo_aretes     = new VALENCES(*topo_aretes);
  topo_quatriemes = new VALENCES(*topo_quatriemes);
  val_quatrieme   = new RE(*val_quatrieme);
  (*refcount) --; refcount = new ulong (1);
}

TOPOLOGIE::TOPOLOGIE (ulong g, ulong d, VALENCES& f, VALENCES& s, VALENCES& a,
   const RE C, const VALENCES q)
: topo_genre (g), topo_degre (d)
, topo_faces (new VALENCES (f))
, topo_sommets (new VALENCES (s))
, topo_aretes (new VALENCES (a))
, topo_quatriemes (new VALENCES (q))
, val_quatrieme (new RE(C))
, refcount (new ulong (1))
{
  long euler = faces().n() + sommets().n() + aretes().n() + quatriemes().n();
  ulong nbv = nombrevaleurs(); do { euler -= degre(); nbv--; } while (nbv > 2);
  if (2 - euler != long(2 * genre())) {
    cerr << "Erreur : la caractristique d'Euler est " << euler
         << " au lieu de " << 2 - 2 * topo_genre
         << " (genre " << topo_genre << ")\n";
    throw lecture;
  }
  ulong i, somme;
  for (i = somme = 0; i < faces().n(); i++) somme += faces()[i];
  if (somme != topo_degre) {
    cerr << "Erreur : la somme des valences des faces est " << somme
         << " au lieu de " << topo_degre << " (degr du dessin)\n";
    throw lecture;
  }
  for (i = somme = 0; i < sommets().n(); i++) somme += sommets()[i];
  if (somme != topo_degre) {
    cerr << "Erreur : la somme des valences des sommets est " << somme
         << " au lieu de " << topo_degre << " (degr du dessin)\n";
    throw lecture;
  }
  for (i = somme = 0; i < aretes().n(); i++) somme += aretes()[i];
  if (somme != topo_degre) {
    cerr << "Erreur : la somme des valences des aretes est " << somme
         << " au lieu de " << topo_degre << " (degr du dessin)\n";
    throw lecture;
  }
  if (nombrevaleurs() == 4) { // Sinon, implicitement tous de valeur 1
    for (i = somme = 0; i < quatriemes().n(); i++) somme += quatriemes()[i];
    if (somme != topo_degre) {
    cerr << "Erreur : la somme des valences des quatrimes est " << somme
         << " au lieu de " << topo_degre << " (degr du dessin)\n";
    throw lecture;
    }
  }
}

ulong
TOPOLOGIE::nombrevaleurs () const
{
  if (quatriemes().n() != 0) return 4;
  else if (faces().n() > 1)  return 3;
  else                       return 2;
}

// ----------------------------------------------------------------------
// class DESSIN : topologie

SELECTION::SELECTION (ulong nbpt, bool possible)
: longueur (nbpt), liste (new ulong[longueur])
{ for (ulong i = 0; i < longueur; i++) liste[i] = possible?absent:impossible; }

SELECTION::SELECTION (const SELECTION& s)
: longueur (s.longueur), liste (new ulong[longueur])
{ for (ulong i = 0; i < longueur; i++) liste[i] = s.liste[i]; }

SELECTION&
SELECTION::operator = (const SELECTION& s)
{
  delete[] liste; longueur = s.longueur; liste = new ulong[longueur];
  for (ulong i = 0; i < longueur; i++) liste[i] = s.liste[i];
  return *this;
}

ulong
SELECTION::nombre () const
{
  ulong n = 0;
  for (ulong i = 0; i < longueur; i++) if (liste[i] == present) n++;
  return n;
}

void
SELECTION::set_impossible (ulong i)
{ if (i<longueur) liste[i] = impossible; }

void
SELECTION::set_possible (ulong i)
{ if (!is_possible(i)) liste[i] = absent; }

void
SELECTION::swap_possible (ulong i)
{ if (is_possible(i)) liste[i] = (liste[i]==absent) ? present : absent; }

bool
SELECTION::is_present (ulong i) const
{ return (i<longueur) ? (liste[i]==present) : false; }

bool
SELECTION::is_possible (ulong i) const
{ return (i<longueur) ? (liste[i]!=impossible) : false; }

bool
SELECTION::same_possible (const SELECTION& s) const
{
  if (longueur != s.longueur) return false;
  for (ulong i = 0; i < longueur; i++)
    if (is_possible(i) != s.is_possible(i))
      return false;
  return true;
}

ANCRE::ANCRE (const ANCRE& a)
{
  type = a.type;
  ancre_nombre = a.ancre_nombre;
  ancre_liste = new ulong [chunk * (ancre_nombre/chunk+1)];
  for (ulong i = 0; i < ancre_nombre; i++) ancre_liste[i] = a.ancre_liste[i];
}

void
ANCRE::add_or_delete (const ulong pt)
{
  switch (type) {
    case nondefini:
         type = point; ancre_nombre = 1; ancre_liste[0] = pt;
         return;
    case point:
         if (pt == ancre_liste[0]) { type = nondefini; ancre_nombre = 0; }
         else { type = somme; ancre_nombre = 2; ancre_liste[1] = pt; }
         return;
    case somme:
    case moyenne:
    case produit:
         for (ulong i = 0; i < ancre_nombre; i++) {
           if (pt == ancre_liste[i]) {
             for (ulong j = i + 1; j < ancre_nombre; j++)
               ancre_liste[j-1] = ancre_liste[j];
             ancre_nombre--; if (ancre_nombre == 1) type = point;
             return;
           }
         }
         ancre_nombre++;
         if ((ancre_nombre+chunk-2)%chunk == 0) {
           ulong* old = ancre_liste;
           ancre_liste = new ulong [chunk * (ancre_nombre/chunk+1)];
           for (ulong j = 0; j < ancre_nombre-1; j++) ancre_liste[j] = old[j];
           delete[] old;
         }
         ancre_liste[ancre_nombre-1] = pt;
         return;
  }
}

void
DESSIN::calcule_ancres ()
{
  ancre0->add_or_delete (1);
  ancre1->add_or_delete (2);
  // TODO : amliorer la recherche des ancres
  const VECTEUR& x = position();
  {for (ulong i = 1; i < x.length(); i++)
    if (IsZero(x[i])) {
      delete ancre0;
      ancre0 = new ANCRE(i);
      cout << "Ancre 0 en 0\n";
      break;
  }}
  {for (ulong i = 1; i < x.length(); i++)
    if (IsOne(x[i])) {
      delete ancre1;
      ancre1 = new ANCRE(i);
      cout << "Ancre 1 en 1\n";
      break;
  }}
}

static ulong
char_count(const char *s, const int c)
{
  ulong n = 0;
  while((s = strchr(s,c))) { n++; s++; }
  return n;
}

VALENCES::VALENCES (const ulong degre)
{
  nombre = degre;
  liste = new ulong[nombre];
  for (ulong i = 0; i < nombre; i++) liste[i] = 1;
}

VALENCES::VALENCES (const VALENCES& a)
{
  nombre = a.nombre;
  liste = new ulong[nombre];
  for (ulong i = 0; i < nombre; i++) liste[i] = a[i];
}

VALENCES::VALENCES (const char* s)
{
  while (*s && (*s==' ')) s++;
  nombre = 1 + char_count(s, ' ');
  liste = new ulong[nombre];
  ulong i, dec;
  for (i = 0; i < nombre; i++) {
    sscanf(s, "%lu %ln", & liste[i], &dec);
    s += dec;
  }
}


// On privilgie la distance minimale deux points, dont l'un de valence minimale
// Sauf si un tel lien existe dj et qu'il n'y a pas de face entre eux
class DISTANCES {
public:
  const ulong nbi; const ulong offi; ulong* vi;
  const ulong nbj; const ulong offj; ulong* vj;
  double* dist;
  ~DISTANCES ();
  DISTANCES (const ulong ni, const ulong oi, const ulong nj, const ulong oj);
  void makedist (const VECTEUR& x);
  void makevals (const VALENCES& ti, const VALENCES& tj);
};
DISTANCES::~DISTANCES ()
{
  if (vi) delete[] vi;
  if (vj) delete[] vj;
  if (dist) delete[] dist;
}
DISTANCES::DISTANCES (const ulong ni, const ulong oi,
                      const ulong nj, const ulong oj)
  : nbi(ni),offi(oi),vi(NULL)
  , nbj(nj),offj(oj),vj(NULL)
  , dist(NULL)
{
  assert (offi>offj);
}
void
DISTANCES::makedist (const VECTEUR& x)
{
  __RE_GARBAGE_BEGIN
  dist = new double[nbi*nbj];
  for (ulong i = 0; i < nbi; i++) for (ulong j = 0; j < nbj; j++)
    dist[i+j*nbi] = double (norm( x[i+offi] - x[j+offj] ));
  __RE_GARBAGE_END
}
void
DISTANCES::makevals (const VALENCES& ti, const VALENCES& tj)
{
  vi = new ulong[nbi]; for(ulong i=0;i<nbi;i++) vi[i] = ti[i];
  vj = new ulong[nbj]; for(ulong j=0;j<nbj;j++) vj[j] = tj[j];
}

static bool
face_entre (const CX& face, const CX& x, const CX& y)
{
  __RE_GARBAGE_BEGIN
  bool ret;
  if ( dot(y-x, face-x) > norm(face-x) ) ret = true; else
  if ( dot(x-y, face-y) > norm(face-y) ) ret = true; else
    ret = false;
  __RE_GARBAGE_END
  return ret;
}

DESSIN::LIEN
plus_court (const ulong nbf, DISTANCES& D, const VECTEUR& x,
            const ulong k, const DESSIN::LIENS& lnks)
{
  // Valence minimale accessible
  ulong v = x.length();
  {for (ulong i=0;i<D.nbi;i++) if (D.vi[i]&&(D.vi[i]<v)) v = D.vi[i];}
  {for (ulong j=0;j<D.nbj;j++) if (D.vj[j]&&(D.vj[j]<v)) v = D.vj[j];}
  if (v == x.length()) return DESSIN::LIEN();
  // Plus courte des distances faisant intervenir cette valence minimale
  ulong l = D.nbj*D.nbi; double d = pInf;
  for(ulong i=0;i<D.nbi;i++) if (D.vi[i])
   for(ulong j=0;j<D.nbj;j++) if (D.vj[j])
    if (D.dist[i+j*D.nbi]<d)  if ((D.vi[i]==v) || (D.vj[j]==v)) {
      // Test si ce lien existe dj et qu'il n'y a pas de face entre eux
      ulong m; for (m = 0; m < k; m++)
        if ((lnks[m].i == i+D.offi)&&(lnks[m].j == j+D.offj)) {
        for (ulong n = 1; n < nbf; n++)
          if (face_entre (x[n], x[i+D.offi], x[j+D.offj]))
            m = k;
        break;
      }
      if ((m == k)||(l == D.nbj*D.nbi)) { l = i+j*D.nbi; d = D.dist[l]; }
  }
  DESSIN::LIEN resultat;
  resultat.i = l % D.nbi; D.vi[resultat.i]--; resultat.i += D.offi;
  resultat.j = l / D.nbi; D.vj[resultat.j]--; resultat.j += D.offj;
  D.dist[l] *= 1.01; // hack pour viter les rptitions nuisibles
  assert (resultat.i>resultat.j);
  return resultat;
}

DESSIN::LIENS::LIENS(DESSIN::LIENS const &a)
{
  cerr << "copy LIENS\n";
  throw this;
}

void
DESSIN::calcule_liens ()
{
  if (liens_fixes) return;
  const VECTEUR& x = position();
  const TOPOLOGIE& t = topologie();
  LIENS& lnks = *dessin_liens;
  if (t.nombrevaleurs() == 4) {
    const ulong nbf=t.faces().n();
    const ulong nbs=t.sommets().n();
    const ulong nba=t.aretes().n();
    const ulong nbq=t.quatriemes().n();
   {
    DISTANCES D (nbq,nbf+nbs+nba, nbs,nbf);
    D. makedist (x);
    D. makevals (t.quatriemes(),t.sommets());
    for (ulong k=0;k<t.degre();k++) {
      lnks[k] = plus_court (nbf, D, x, k, lnks);
      if (lnks[k].i==lnks[k].j) break;
    }
   } 
   {
    DISTANCES D (nbq,nbf+nbs+nba, nba,nbf+nbs);
    D. makedist (x);
    D. makevals (t.quatriemes(),t.aretes());
    for (ulong k=t.degre();k<2*t.degre();k++) {
      lnks[k] = plus_court (nbf, D, x, k, lnks);
      if (lnks[k].i==lnks[k].j) break;
    }
   }
  }
  else {
    const ulong nbf=t.faces().n();
    const ulong nbs=t.sommets().n();
    const ulong nba=t.aretes().n();
    DISTANCES D (nba,nbf+nbs, nbs,nbf);
    D. makedist (x);
    D. makevals (t.aretes(),t.sommets());
    for (ulong k=0;k<t.degre();k++) {
      lnks[k] = plus_court (nbf, D, x, k, lnks);
      if (lnks[k].i==lnks[k].j) break;
    }
  }
}

void
DESSIN::fixeliens ()
{
  liens_fixes = ! liens_fixes;
}

void
DESSIN::demiliens ()
{
  switch (liens_morceaux) {
    case  0: liens_morceaux = 1; break;
    case  1: liens_morceaux = 2; break;
    case  2: liens_morceaux = 3; break;
    case  3: liens_morceaux = 5; break;
    case  5: liens_morceaux = 8; break;
    case  8: liens_morceaux = 13; break;
    default: liens_morceaux = 0; break;
  }
  if (topologie().nombrevaleurs() == 4) liens_morceaux = 0;
}

static void
reordonne (const VECTEUR& base, VECTEUR& voisins)
{
  const ulong nb = base.length();
  assert (nb == voisins.length());
  // Calcul des distances
  double dist[nb][nb];
  __RE_GARBAGE_BEGIN
  {for (ulong i = 0; i < nb; i++)
    for (ulong j = 0; j < nb; j++)
      dist[j][i] = double (norm (base[i] - voisins[j]));}
  __RE_GARBAGE_END
  for (ulong k = 0; k < nb; k++) {
    // Recherche de la plus courte
    double d = pInf; ulong ip, jp;
    for (ulong i = 0; i < nb; i++)
      if (dist[i][i] < d) { d = dist[i][i]; ip = i; jp = i;}
    for (ulong i = 0; i < nb; i++) {
      for (ulong j = 0; j < nb; j++)
        if (dist[j][i] * 1.01 < d) { d = dist[j][i]; ip = i; jp = j;}
    }
    for (ulong i = 0; i < nb; i++) dist[jp][i] = pInf;
    for (ulong j = 0; j < nb; j++) dist[j][ip] = pInf;
    // ventuellement swap
    if (ip != jp) {
      double x[nb]; x = dist[jp]; dist[jp] = dist[ip]; dist[ip] = x;
      CX v = voisins[jp]; voisins[jp] = voisins[ip]; voisins[ip] = v;
    }
  }
}

void
DESSIN::demi_positions ()
{
  assert (topologie().nombrevaleurs() != 4);
  if (liens0p && liens0q && (liens_morceaux==0)) return;
  if (liens1 && (liens_morceaux<=1)) return;
  if (liens2p && liens2q && (liens_morceaux<=2)) return;
  if (liens3p && liens3q && (liens_morceaux<=3)) return;
  if (liens5p && liens5q && (liens_morceaux<=5)) return;
  if (liens8p && liens8q && (liens_morceaux<=8)) return;
  if (liens13p && liens13q && (liens_morceaux<=13)) return;
  const BELYI& b = belyi();
  LIENS& lnks = *dessin_liens;
  liens0p = new VECTEUR (lnks.nb);
  for (ulong k = 0; k < lnks.nb; k++) {
    if (lnks[k].i <= lnks[k].j) { cerr << "Liens non valides\n"; throw this; }
    (*liens0p)[k] = position()[lnks[k].j];
  }
  liens0q = new VECTEUR (lnks.nb);
  for (ulong k = 0; k < lnks.nb; k++) {
    if (lnks[k].i <= lnks[k].j) { cerr << "Liens non valides\n"; throw this; }
    (*liens0q)[k] = position()[lnks[k].i];
  }
  VECTEUR v = *liens0p; v += *liens0q; v /= 2;
  if (liens_morceaux>=1) {
    liens1 = new VECTEUR (racines (b.P+b.Q, v));
    reordonne (*liens0p, *liens1);
  }
  if (liens_morceaux>=2) {
    v = *liens0p; v += *liens1; v /= 2;
    liens2p = new VECTEUR (racines (3*b.P+b.Q, v));
    reordonne (*liens1, *liens2p);
    v = *liens0q; v += *liens1; v /= 2;
    liens2q = new VECTEUR (racines (b.P+3*b.Q, v));
    reordonne (*liens1, *liens2q);
  }
  if (liens_morceaux>=3) {
    v = *liens0p; v += *liens2p; v /= 2;
    liens3p = new VECTEUR (racines (7*b.P+b.Q, v));
    reordonne (*liens2p, *liens3p);
    v = *liens0q; v += *liens2q; v /= 2;
    liens3q = new VECTEUR (racines (b.P+7*b.Q, v));
    reordonne (*liens2q, *liens3q);
  }
  if (liens_morceaux>=5) {
    v = *liens0p; v += *liens3p; v /= 2;
    liens5p = new VECTEUR (racines (31*b.P+b.Q, v));
    reordonne (*liens3p, *liens5p);
    v = *liens0q; v += *liens3q; v /= 2;
    liens5q = new VECTEUR (racines (b.P+31*b.Q, v));
    reordonne (*liens3q, *liens5q);
  }
  if (liens_morceaux>=8) {
    v = *liens0p; v += *liens5p; v /= 2;
    liens8p = new VECTEUR (racines (255*b.P+b.Q, v));
    reordonne (*liens5p, *liens8p);
    v = *liens0q; v += *liens5q; v /= 2;
    liens8q = new VECTEUR (racines (b.P+255*b.Q, v));
    reordonne (*liens5q, *liens8q);
  }
  if (liens_morceaux>=13) {
    v = *liens0p; v += *liens8p; v /= 2;
    liens13p = new VECTEUR (racines (8191*b.P+b.Q, v));
    reordonne (*liens8p, *liens13p);
    v = *liens0q; v += *liens8q; v /= 2;
    liens13q = new VECTEUR (racines (b.P+8191*b.Q, v));
    reordonne (*liens8q, *liens13q);
  }
}

const char*
TOPOLOGIE::variable (ulong i) const
{
  if (i == 0)               return "lambda";
  static char S[10];
  if (i < faces().n())      { snprintf(S, 10, "r%ld", i); return S; }
  i -= faces().n();
  if (i < sommets().n())    { snprintf(S, 10, "p%ld", i); return S; }
  i -= sommets().n();
  if (i < aretes().n())     { snprintf(S, 10, "q%ld", i); return S; }
  i -= aretes().n();
  if (i < quatriemes().n()) { snprintf(S, 10, "s%ld", i); return S; }
  return "ERREUR";
}

TOPOLOGIE::type
TOPOLOGIE::typeramif (ulong i) const
{
  if (i == 0)               return lambda;
  if (i < faces().n())      return face;   i -= faces().n();
  if (i < sommets().n())    return sommet; i -= sommets().n();
  if (i < aretes().n())     return arete;  i -= aretes().n();
  if (i < quatriemes().n()) return quatrieme;
  return lambda;
}

ulong
TOPOLOGIE::valence (ulong i) const
{
  if (i < faces().n())      return faces()[i];   i -= faces().n();
  if (i < sommets().n())    return sommets()[i]; i -= sommets().n();
  if (i < aretes().n())     return aretes()[i];  i -= aretes().n();
  if (i < quatriemes().n()) return quatriemes()[i];
  return 0;
}

ulong
DESSIN::voisin (const CX& c) const
{
  double mindist = pInf;
  __RE_GARBAGE_BEGIN
  ulong j; for (ulong i = 1; i < position().length(); i++) {
    double dist = norm( position()[i] - c );
    if (dist < mindist) { mindist = dist; j = i; }
  }
  __RE_GARBAGE_END
  return j;
}

void
DESSIN::info (const CX& c) const
{
  cout << IOmedium << "Click en " << c << "\n";
  ulong v = voisin (c);
  cout << topologie().variable(v) << ", n " << v << ", valence "
       << topologie().valence(v) << " : " << position()[v] << "\n";
}

// ----------------------------------------------------------------------
// class DESSIN : constructeur/destructeur/modificateur/io

DESSIN::~DESSIN ()
{
  delete dessin_topo;
  delete dessin_pos;
  delete dessin_liens;
  delete ancre0;
  delete ancre1;
  if (liens0p)         delete liens0p;
  if (liens0q)         delete liens0q;
  if (liens1)          delete liens1;
  if (liens2p)         delete liens2p;
  if (liens2q)         delete liens2q;
  if (liens3p)         delete liens3p;
  if (liens3q)         delete liens3q;
  if (liens5p)         delete liens5p;
  if (liens5q)         delete liens5q;
  if (liens8p)         delete liens8p;
  if (liens8q)         delete liens8q;
  if (liens13p)        delete liens13p;
  if (liens13q)        delete liens13q;
  if (dessin_belyi)    delete dessin_belyi;
  if (dessin_negimage) delete dessin_negimage;
  if (dessin_derivee)  delete dessin_derivee;
}

DESSIN::DESSIN (const char* filename)
: historique (50)
{
  FILEREAD infile(filename);
  if (!infile) throw lecture;
  // ---------------------------------------------------
  // Lecture des donnes du dessin : topologie, position
  const char* en_tete = infile.scanline();
  ulong en_tete_version = 0;
  if (en_tete[0] == '<') {
    cout << "Fichier de donnes de type " << en_tete << "\n";
    en_tete_version ++;
    en_tete = infile.scanline();
  }
  ulong genre, degre, nbramif = 0;
  sscanf(en_tete,"%lu %lu %lu", &genre, &degre, &nbramif);
  switch (nbramif) {
    case 2: {
      char facevalence[10]; sprintf(facevalence,"%ld",degre);
      VALENCES faces     (facevalence);
      VALENCES sommets   (infile.scanline());
      VALENCES aretes    (infile.scanline());
      dessin_topo  = new TOPOLOGIE(genre, degre, faces, sommets, aretes);
      dessin_pos   = new VECTEUR(degre+2);
      for (ulong i = 1; i < degre+2; i++)
        (*dessin_pos)[i] = infile.scanline();
      } break;
    case 3: {
      VALENCES faces     (infile.scanline());
      VALENCES sommets   (infile.scanline());
      VALENCES aretes    (infile.scanline());
      dessin_topo  = new TOPOLOGIE(genre, degre, faces, sommets, aretes);
      dessin_pos   = new VECTEUR(degre+2);
      for (ulong i = 0; i < degre+2; i++)
        (*dessin_pos)[i] = infile.scanline();
      } break;
    case 4: {
      VALENCES faces     (infile.scanline());
      VALENCES sommets   (infile.scanline());
      VALENCES aretes    (infile.scanline());
      VALENCES quatrieme (infile.scanline());
      RE C = infile.scanline();
      dessin_topo  = new TOPOLOGIE(genre, degre, faces, sommets, aretes,
         C, quatrieme);
      dessin_pos   = new VECTEUR(2*degre+2);
      for (ulong i = 0; i < 2*degre+2; i++)
        (*dessin_pos)[i] = infile.scanline();
      } break;
    case 0:
      if (en_tete_version != 0) {
        cerr <<
        "Le format devrait prciser le nombre de valeurs de ramification\n";
        throw lecture;
      } else {
      VALENCES faces     (infile.scanline());
      VALENCES sommets   (infile.scanline());
      VALENCES aretes    (infile.scanline());
      dessin_topo  = new TOPOLOGIE(genre, degre, faces, sommets, aretes);
      dessin_pos   = new VECTEUR(degre+2);
      ulong i = 0;
      if (dessin_topo->faces().n()==1) { i++;cout<<"Compatibilit - arbres\n"; }
      for (; i < degre+2; i++)
        (*dessin_pos)[i] = infile.scanline();
      } break;
    default:
      cerr << "Nombre de valeurs de ramification invalide : " <<nbramif<< "\n";
      throw lecture;
  }
  dessin_liens    = new LIENS (topologie());
  liens_fixes     = false;
  liens0p         = NULL;
  liens0q         = NULL;
  liens1          = NULL;
  liens2p         = NULL;
  liens2q         = NULL;
  liens3p         = NULL;
  liens3q         = NULL;
  liens5p         = NULL;
  liens5q         = NULL;
  liens8p         = NULL;
  liens8q         = NULL;
  liens13p        = NULL;
  liens13q        = NULL;
  liens_morceaux  = 0;
  ancre0          = new ANCRE ();
  ancre1          = new ANCRE ();
  dessin_belyi    = NULL;
  dessin_negimage = NULL;
  dessin_derivee  = NULL;
  // ---------------------------------------------------
  // Calculs annexes
  calcule_ancres();
  // calcul de x[0] pour les arbres
  if (topologie().nombrevaleurs() == 2) ajustelambda (false);
}

void
DESSIN::chargement (const char* filename)
{
  try {
  DESSIN d (filename);
  empile ();
#define swapfield(type,champ) { type* X=champ; champ=d.champ; d.champ=X; }
  swapfield (TOPOLOGIE, dessin_topo)
  swapfield (VECTEUR,   dessin_pos)
  swapfield (LIENS,     dessin_liens)
  swapfield (BELYI,     dessin_belyi)
  swapfield (ANCRE,     ancre0)
  swapfield (ANCRE,     ancre1)
  swapfield (VECTEUR,   dessin_negimage)
  swapfield (MCARREE,   dessin_derivee)
#undef swapfield
  liens_fixes     = false;
  liens0p = NULL; if (liens0p) delete liens0p;
  liens0q = NULL; if (liens0q) delete liens0q;
  liens1  = NULL; if (liens1)  delete liens1;
  liens2p = NULL; if (liens2p) delete liens2p;
  liens2q = NULL; if (liens2q) delete liens2q;
  liens3p = NULL; if (liens3p) delete liens3p;
  liens3q = NULL; if (liens3q) delete liens3q;
  liens5p = NULL; if (liens5p) delete liens5p;
  liens5q = NULL; if (liens5q) delete liens5q;
  liens8p = NULL; if (liens8p) delete liens8p;
  liens8q = NULL; if (liens8q) delete liens8q;
  liens13p = NULL; if (liens13p) delete liens13p;
  liens13q = NULL; if (liens13q) delete liens13q;
  liens_morceaux  = 0;
  }
  catch (exceptions lecture) { }
}

ostream&
operator << (ostream& s, const TOPOLOGIE& t)
{
  s << "# Genre, Degre, Nombre de type de points\n";
  s << " " << t.genre() << " " << t.degre() << " " << t.nombrevaleurs() << "\n";
  if (t.nombrevaleurs() > 2) {
    s << "# R (faces)\n";
    for (ulong i = 0; i < t.faces().n(); i++) s <<" "<< t.faces()[i];
    s << "\n";
  }
  s << "# P (sommets)\n";
  {for (ulong i = 0; i < t.sommets().n(); i++) s <<" "<< t.sommets()[i];}
  s << "\n";
  s << "# Q (aretes)\n";
  {for (ulong i = 0; i < t.aretes().n(); i++) s <<" "<< t.aretes()[i];}
  s << "\n";
  if (t.nombrevaleurs() == 4) {
    s << "# S (quatriemes)\n";
    for (ulong i = 0; i < t.quatriemes().n(); i++) s <<" "<< t.quatriemes()[i];
    s << "\n";
    s << "# C (quatrime valeur)\n";
    s << IOlong << t.C() << "\n";
  }
  return s;
}

ostream&
operator << (ostream& s, const ANCRE& a)
{
  switch (a.type) {
    case ANCRE::nondefini:   s << "Non dfinie"; break;
    case ANCRE::point:       s << "Point"; break;
    case ANCRE::somme:       s << "Somme"; break;
    case ANCRE::moyenne:     s << "Moyenne"; break;
    case ANCRE::produit:     s << "Produit"; break;
  }
  for (ulong j = 0; j < a.nombre(); j++) s << " " << a[j];
  return s;
}

ostream&
operator << (ostream& s, const DESSIN& d)
{
  const TOPOLOGIE& t = d.topologie();
  const VECTEUR& x = d.position();
  s << "<!DESSIN v1>\n" << t;
  s << "# Ancre 0 : " << *d.ancre0 << "\n";
  s << "# Ancre 1 : " << *d.ancre1 << "\n";
  s << "# Points\n" << IOlong;
  if (t.nombrevaleurs() > 2)
    s << " " << x[0] << " lambda\n";
  for (ulong i = 1; i < x.length(); i++)
    s << " " << x[i] << " " << t.variable(i) << "\n";
  return s;
}

void
DESSIN::moved ()
{
  liens0p = NULL; if (liens0p) delete liens0p;
  liens0q = NULL; if (liens0q) delete liens0q;
  liens1  = NULL; if (liens1)  delete liens1;
  liens2p = NULL; if (liens2p) delete liens2p;
  liens2q = NULL; if (liens2q) delete liens2q;
  liens3p = NULL; if (liens3p) delete liens3p;
  liens3q = NULL; if (liens3q) delete liens3q;
  liens5p = NULL; if (liens5p) delete liens5p;
  liens5q = NULL; if (liens5q) delete liens5q;
  liens8p = NULL; if (liens8p) delete liens8p;
  liens8q = NULL; if (liens8q) delete liens8q;
  liens13p = NULL; if (liens13p) delete liens13p;
  liens13q = NULL; if (liens13q) delete liens13q;
  if (dessin_belyi)    delete dessin_belyi;    dessin_belyi = NULL;
  if (dessin_negimage) delete dessin_negimage; dessin_negimage = NULL;
  if (dessin_derivee)  delete dessin_derivee;  dessin_derivee = NULL;
}

void
DESSIN::unsplit ()
{
  if (nombrevaleurs() != 4) throw this;
  if (!IsZero(topologie().C()) && !IsOne(topologie().C()))
  {
    cerr << "Erreur : pour fusionner, la quatrime valeur doit valoir 0 ou 1\n"
         << "         et non pas " << topologie().C() << "\n";
    return;
  }
  // Recherche de l'unique point d'ordre 2 au dessus de C
  const VECTEUR& x = position();
  const TOPOLOGIE& t = topologie();
  const ulong nbf=t.faces().n();
  const ulong nbs=t.sommets().n();
  const ulong nba=t.aretes().n();
  const ulong nbq=t.quatriemes().n();
  ulong Cpos = 0;
  const VALENCES& q = t.quatriemes();
  {for (ulong i = 0; i < nbq; i++) switch (q[i]) {
    case 1: break;
    case 2: if (!Cpos) { Cpos = nbf+nbs+nba+i; break; }
    default: cerr << "Erreur : il faut qu'il y ait un unique point"
                  << " d'ordre 2 au dessus de C\n";
  }}
  const CX& Cval = x[Cpos];
  ulong X = 0, Y = 0; // deux points a fusionner
  double d0 = pInf; double d1 = pInf;
  if (IsZero(topologie().C())) {
    for (ulong i = nbf; i < nbf+nbs; i++) {
      double d = norm( Cval - x[i] );
      if (d < d0) { Y = X; X = i; d1 = d0; d0 = d; }
      else if (d < d1) { Y = i; d1 = d; }
    }
  }
  if (IsOne(topologie().C())) {
    for (ulong i = nbf+nbs; i < nbf+nbs+nba; i++) {
      double d = norm( Cval - x[i] );
      if (d < d0) { Y = X; X = i; d1 = d0; d0 = d; }
      else if (d < d1) { Y = i; d1 = d; }
    }
  }
  empile ();
  if (Y < X) { ulong Z = X; X = Y; Y = Z; }
  VECTEUR* pos = new VECTEUR(t.degre()+2);
  {for (ulong i = 0; i < Y; i++)               (*pos)[i] = x[i];}
  {for (ulong i = Y; i < (*pos).length(); i++) (*pos)[i] = x[i+1];}
  (*pos)[X] = Cval;
  delete dessin_pos;   dessin_pos = pos;
  dessin_topo -> fusion (X, Y);
  ancre0      -> fusion (X, Y, t.degre()+2);
  ancre1      -> fusion (X, Y, t.degre()+2);
  delete dessin_liens; dessin_liens = new LIENS (topologie());
  liens_fixes     = false;
  // TODO : calculer nouveaux liens (si liens_fixe)
  moved ();
}

void
ANCRE::fusion (ulong i, ulong j, ulong max)
{
  {for (ulong k = 0; k < ancre_nombre; k++)
    if (ancre_liste[k] == j)
      ancre_liste[k] = i;
    else if (ancre_liste[k] > j)
      ancre_liste[k] --;
  }
  ulong decale = 0;
  {for (ulong k = 0; k < ancre_nombre; k++) {
    if (ancre_liste[k] >= max) decale++;
    ancre_liste[k] = ancre_liste[k+decale];
  }}
  ancre_nombre -= decale;
  if      (ancre_nombre == 0) type = nondefini;
  else if (ancre_nombre == 0) type = point;
}

void
TOPOLOGIE::fusion (ulong i, ulong j)
{
  const ulong nbf=faces().n();
  const ulong nbs=sommets().n();
  topo_faces      = new VALENCES(*topo_faces);
  topo_sommets    = new VALENCES(*topo_sommets);
  topo_aretes     = new VALENCES(*topo_aretes);
  topo_quatriemes = new VALENCES();
  if (IsZero(C())) {
    i -= nbf; j -= nbf;
    topo_sommets -> fusion (i, j);
  }
  if (IsOne(C())) {
    i -= nbf + nbs; j -= nbf + nbs;
    topo_aretes  -> fusion (i, j);
  }
  val_quatrieme = new RE();
  refcount = new ulong (1);
}

void
VALENCES::fusion (ulong i, ulong j)
{
  if ((i >= nombre) || (j >= nombre))
  { cerr << "VALENCES , fusion out of range\n"; return; }
  if (j < i) { ulong k = i; i = j; j = k; }
  liste[i] += liste[j];
  nombre--;
  for (ulong k = j; k < nombre; k++)
    liste[k] = liste[k+1];
}

VECTEUR*
DESSIN::liste_voisins (const ulong mark)
{
  bool is_sommet = topologie().typeramif(mark) == TOPOLOGIE::sommet;
  switch (liens_morceaux) {
    case 0:  return is_sommet ? liens0q : liens0p; break;
    case 1:  return liens1; break;
    case 2:  return is_sommet ? liens2p : liens2q; break;
    case 3:  return is_sommet ? liens3p : liens3q; break;
    case 5:  return is_sommet ? liens5p : liens5q; break;
    case 8:  return is_sommet ? liens8p : liens8q; break;
    case 13: return is_sommet ? liens13p : liens13q; break;
  }
  throw this;
  return NULL;
}

SELECTION
DESSIN::voisins (const ulong mark)
{
  assert (nombrevaleurs() != 4);
  assert (mark > 0);
  const TOPOLOGIE& t = topologie();
  assert (t.valence(mark) >= 2);
  const LIENS& lnks = *dessin_liens;
  SELECTION out (t.degre(), false);
  if (t.typeramif(mark) == TOPOLOGIE::sommet) {
    for (ulong k = 0; k < t.degre(); k++)
      if (lnks[k].j==mark) out.set_possible(k);
  }
  else {
    for (ulong k = 0; k < t.degre(); k++)
      if (lnks[k].i==mark) out.set_possible(k);
  }
  return out;
}

void
DESSIN::split (ulong mark, SELECTION selection)
{
  assert (nombrevaleurs() != 4);
  assert (selection.nombre () == 2);
  const VECTEUR& x = position();
  const TOPOLOGIE& t = topologie();
  const ulong nbf=t.faces().n();
  const ulong nbs=t.sommets().n();
  const ulong nba=t.aretes().n();
  calcule_liens ();
  demi_positions ();
  const LIENS& lnks = *dessin_liens;
  ulong X = 0;   for (; X < t.degre(); X++) if (selection.is_present (X)) break;
  ulong Y = X+1; for (; Y < t.degre(); Y++) if (selection.is_present (Y)) break;
  if (t.typeramif (mark) == TOPOLOGIE::sommet) {
    cout << "Sont slectionns " << t.variable(lnks[X].i);
    cout << " et " << t.variable(lnks[Y].i) << "\n";
  }
  else {
    cout << "Sont slectionns " << t.variable(lnks[X].j);
    cout << " et " << t.variable(lnks[Y].j) << "\n";
  }
  
  VECTEUR* pos_fin = liste_voisins (mark);
  //CX ligne = (*pos_fin)[Y] - (*pos_fin)[X];
  CX ligne = unit((*pos_fin)[Y]-x[mark]) - unit((*pos_fin)[X]-x[mark]);
  ligne = unit(ligne) * norm((*pos_fin)[Y]-(*pos_fin)[X]);
  ligne *= CX(0,1);
  if (IsZero(ligne)) { ligne = x[mark]-(*pos_fin)[X]; cout<<"Lien multiple\n"; }
  // x[mark] sera au dessus de la ligne X...Y et x[newpoint] au dessous
  ulong dessous = 1;
  for (ulong k = 0; k < t.degre(); k++) {
    if (!selection.is_possible (k)) continue;
    if ((k==X)||(k==Y)) continue;
    if (dot(ligne, (*pos_fin)[k]-x[mark]) < RE(0)) dessous ++;
  }
  cout << "Le point " << t.variable(mark) << " de valence "
       << t.valence(mark) << " est splitt en deux points de valences "
       << t.valence(mark) - dessous << " et " << dessous << ".\n";
  // TODO : choix de epsilon
  const double epsilon = 1e-2;
  cout << "epsilon vaut " << epsilon << "\n";
  
  empile ();
  ulong newpoint, qpoint;
  VECTEUR* pos = new VECTEUR(2*t.degre()+2);
  switch (t.typeramif(mark)) {
    case TOPOLOGIE::sommet:
      // Les points qui ne bougent pas
      {for (ulong i = 0; i < nbf+nbs; i++) (*pos)[i] = x[i];}
      {for (ulong i = nbf+nbs; i < nbf+nbs+nba; i++) (*pos)[i+1] = x[i];}
      // Nouveau point et point marqu
      newpoint = nbf+nbs;
      (*pos)[nbf+nbs+nba+1] = x[mark];
      (*pos)[newpoint] = x[mark] - epsilon/2 * ligne;
      (*pos)[mark]     = x[mark] + epsilon/2 * ligne;
      // Les nouvelles ramifications
      qpoint = nbf+nbs+nba+2;
#if 1
      {for (ulong k = 0; k < t.degre(); k++) {
        if (lnks[k].j!=mark) continue;
        if ((k==X)||(k==Y)) continue;
        if (dot(ligne, (*pos_fin)[k]-x[mark]) < RE(0))
          (*pos)[qpoint++] = (*pos)[newpoint]
                           + epsilon * ((*pos_fin)[k]-(*pos)[newpoint]);
        else
          (*pos)[qpoint++] = (*pos)[mark]
                           + epsilon * ((*pos_fin)[k]-(*pos)[mark]);
      }}
#if 0
      {for (ulong v = t.valence(mark) - dessous; v > 1; v--)
        (*pos)[qpoint++] = (*pos)[mark]; // TODO + decalage
      }
      {for (ulong v = dessous; v > 1; v--)
        (*pos)[qpoint++] = (*pos)[newpoint]; // TODO + decalage
      }
#endif
      {for (ulong k = 0; k < nbs; k++) {
        if (mark==k+nbf) continue;
        const CX& xpos = x[k+nbf];
        const ulong val = t.valence (k+nbf);
        // Calcul de A x^v = P(x)/lambdaR(x)
        CX A = 1;
        for (ulong i = 0; i < nbs; i++) {
          if (i==k) continue;
          for (ulong v = t.valence (i+nbf); v > 0; v--)
            A *= xpos - x[i+nbf];
        }
        for (ulong i = 1; i < nbf; i++)
          for (ulong v = t.valence (i); v > 0; v--) A /= xpos - x[i];
        A /= x[0];
        A = root (epsilon/A, val);
        for (ulong v = val; v > 0; v--)
          (*pos)[qpoint++] = xpos + A * direction((0.+v)/val);
      }}
#else
      {for (ulong k = 0; k < t.degre(); k++) {
        if ((k==X)||(k==Y)) continue;
        const CX* thispoint = & x[lnks[k].j];
        if (lnks[k].j==mark)
          thispoint = & (*pos)[
            dot(ligne, x[lnks[k].i]-x[mark]) < RE(0)
            ? newpoint : mark
            ];
        (*pos)[qpoint++] = *thispoint + epsilon * (x[lnks[k].i]-*thispoint);
      }}
#endif
      break;
    case TOPOLOGIE::arete:
      // Les points qui ne bougent pas
      {for (ulong i = 0; i < nbf+nbs+nba; i++) (*pos)[i] = x[i];}
      // Nouveau point et point marqu
      newpoint = nbf+nbs+nba;
      (*pos)[nbf+nbs+nba+1] = x[mark];
      (*pos)[newpoint] = x[mark] - epsilon/2 * ligne;
      (*pos)[mark]     = x[mark] + epsilon/2 * ligne;
      // Les nouvelles ramifications
      qpoint = nbf+nbs+nba+2;
      {for (ulong k = 0; k < t.degre(); k++) {
        if ((k==X)||(k==Y)) continue;
        const CX* thispoint = & x[lnks[k].i];
        if (lnks[k].i==mark)
          thispoint = & (*pos)[
            dot(ligne, x[lnks[k].j]-x[mark]) < RE(0)
            ? newpoint : mark
            ];
        (*pos)[qpoint++] = *thispoint + epsilon * (x[lnks[k].j]-*thispoint);
      }}
      break;
    default:
      cerr << "ERROR: Never Reached\n";
      throw this;
  }
  delete dessin_pos;   dessin_pos = pos;
  dessin_topo -> split (mark, dessous);
  ancre0      -> split (newpoint);
  ancre1      -> split (newpoint);
  delete dessin_liens; dessin_liens = new LIENS (topologie());
  liens_fixes     = false;
  liens_morceaux  = 0;
  moved ();
}

void
TOPOLOGIE::split (ulong i, ulong v)
{
  const ulong nbf=faces().n();
  const ulong nbs=sommets().n();
  topo_faces      = new VALENCES(*topo_faces);
  topo_sommets    = new VALENCES(*topo_sommets);
  topo_aretes     = new VALENCES(*topo_aretes);
  topo_quatriemes = new VALENCES(degre());
  switch (typeramif(i)) {
    case TOPOLOGIE::sommet:
      topo_sommets    -> split (i - nbf, v);
      topo_quatriemes -> fusion (0, 1);
      val_quatrieme    = new RE(0);
      break;
    case TOPOLOGIE::arete:
      topo_aretes     -> split (i - nbf - nbs, v);
      topo_quatriemes -> fusion (0, 1);
      val_quatrieme    = new RE(1);
      break;
    default:
      cerr << "ERROR: Never Reached\n";
      throw this;
  }
  refcount = new ulong (1);
}

void
VALENCES::split (ulong i, ulong v)
{
  if (i >= nombre)
  { cerr << "VALENCES , split out of range\n"; return; }
  ulong* l = new ulong[nombre+1];
  for (ulong k = 0; k < nombre; k++) l[k] = liste[k];
  l[i] -= v; l[nombre] = v;
  delete[] liste; liste = l; nombre++;
}

void
ANCRE::split (ulong newpoint)
{
  for (ulong k = 0; k < ancre_nombre; k++)
    if (ancre_liste[k] >= newpoint)
      ancre_liste[k] ++;
}

void
TOPOLOGIE::moveC (const RE& newC)
{
  if (nombrevaleurs() != 4) throw this;
  copy_if_multiref();
  *val_quatrieme = newC;
  if (C() < RE(0)) {cerr << "C doit tre positif\n"; *val_quatrieme = 0; }
  if (C() > RE(1)) {cerr << "C doit tre infrieur  1\n"; *val_quatrieme = 1; }
}

void
DESSIN::moveC (const RE& newC)
{
  empile ();
  dessin_topo->moveC(newC);
  moved ();
}

void
DESSIN::move (const VECTEUR& decalage)
{
  empile ();
  *dessin_pos += decalage;
  moved ();
}

void
DESSIN::undo ()
{
  depile ();
  delete dessin_liens; dessin_liens = new LIENS (topologie());
  liens_fixes     = false;
  moved ();
}

void
DESSIN::move_ancre (const ulong mark, ANCRE& a)
{
  if (mark) { empile (); a.add_or_delete(mark); moved (); return; }
  switch (a.type) {
    case ANCRE::nondefini:
    case ANCRE::point:
               break;
    case ANCRE::somme:   empile (); a.type = ANCRE::moyenne; moved (); break;
    // TODO: produit... le calcul foire pour le moment
    case ANCRE::moyenne: //empile (); a.type = ANCRE::produit; moved (); break;
    case ANCRE::produit: empile (); a.type = ANCRE::somme;   moved (); break;
  }
}

void
DESSIN::ajusteancres ()
{
  CX val0 = negvaleur (*ancre0, position());
  if (!IsZero(val0)) {
    empile ();
    for (ulong i = 1; i < position().length(); i++)
      (*dessin_pos)[i] += val0;
  }
  CX val1 = negvaleur (*ancre1, position());
  if (!IsZero(val1)) {
    if (IsZero(val0)) empile ();
    *dessin_pos /= -val1;
    for (ulong i = 1; i < topologie().valence(0); i++)
      (*dessin_pos)[0] /= -val1;
  }
  if (!IsZero(val0) || !IsZero(val1)) moved ();
}

void
DESSIN::ajustelambda (bool empiler)
{
  // Lambda = (P-Q) / R
  const BELYI& b = belyi ();
  ulong deg = b.R.degre();
  CX newlambda = b.P[deg] - b.Q[deg];
  if (newlambda != (*dessin_pos)[0]) {
    if (empiler) empile ();
    (*dessin_pos)[0] = newlambda;
    moved ();
  }
}

static void
pos_inversion (VECTEUR& x, ulong mark, ulong limite, ulong decalage)
{
  VECTEUR pos = x;
  x[decalage] = 0;
  {for (ulong i = 1; i < limite; i++)   x[i+decalage] = pos[i];}
  x[0] = pos[0];
  {for (ulong i = 1; i < decalage; i++) x[i] = pos[i+limite];}
  if (mark != limite)
    x[mark-limite] = pos[limite];
}

static void
ancre_inversion (ANCRE*& a, ulong mark, ulong limite, ulong decalage)
{
  ANCRE* a1 = a; a = new ANCRE();
  for (ulong i = 0; i < a1->nombre(); i++) {
    ulong x = (*a1)[i]; ulong y;
    if (x < limite) y = x + decalage;
    else y = x - limite;
    if (x == limite) y = (mark<limite) ? mark+decalage : mark-limite;
    if (x == mark) y = decalage;
    a->add_or_delete (y);
  } a->type = a1->type;
  delete a1;
}

void
DESSIN::inversion (const ulong mark)
{
  if (nombrevaleurs() == 4) {
    cerr << "TODO: inversion, quatre valeurs\n";
    return;
  }
  empile ();
  // Translation : mark -> 0
  const CX posmark = (*dessin_pos)[mark];
  {for (ulong i = 1; i < position().length(); i++)
    (*dessin_pos)[i] -= posmark;}
  // Ajustement de lambda
  const TOPOLOGIE& t = topologie();
  const ulong nbf = t.faces().n();
  const ulong nbs = t.sommets().n();
  const ulong nba = t.aretes().n();
  CX prodR = 1; {for (ulong i = 1; i < nbf; i++) if (i != mark)
    for (ulong j = 0; j < t.valence(i); j++) prodR *= - (*dessin_pos)[i];}
  CX prodP = 1; {for (ulong i = nbf; i < nbf+nbs; i++) if (i != mark)
    for (ulong j = 0; j < t.valence(i); j++) prodP *= - (*dessin_pos)[i];}
  CX prodQ = 1; {for (ulong i = nbf+nbs; i < nbf+nbs+nba; i++) if (i != mark)
    for (ulong j = 0; j < t.valence(i); j++) prodQ *= - (*dessin_pos)[i];}
  switch (t.typeramif(mark)) {
    case TOPOLOGIE::face:
      // cout << "Prod(-r) = " << prodR << "\n";
      // cout << "Prod(-p)  = " << prodP << "\n";
      // cout << "Prod(-q)  = " << prodQ << "\n";
      (*dessin_pos)[0] *= prodR / prodP;
      break;
    case TOPOLOGIE::sommet:
      // cout << "Prod(-p)           = " << prodP << "\n";
      // cout << "Prod(-q)            = " << prodQ << "\n";
      // cout << "- Lambda * Prod(-r) = " << - (*dessin_pos)(0) * prodR << "\n";
      (*dessin_pos)[0] = prodP / prodQ;
      break;
    case TOPOLOGIE::arete:
      // cout << "- Prod(-q)       = " << - prodQ << "\n";
      // cout << "Prod(-p)          = " << prodP << "\n";
      // cout << "Lambda * Prod(-r) = " << (*dessin_pos)(0) * prodR << "\n";
      (*dessin_pos)[0] = - prodQ / prodP;
      break;
    case TOPOLOGIE::quatrieme:
    case TOPOLOGIE::lambda:
      cerr << "ERROR: Never Reached\n";
      throw this;
  }
  // Inversion (sauf mark)
  {for (ulong i = 1; i < position().length(); i++)
    if (i != mark)
    (*dessin_pos)[i] = inv ( (*dessin_pos)[i] );}
  // Changement des valences
  switch (t.typeramif(mark)) {
    case TOPOLOGIE::face:
      // c'est OK
      break;
    case TOPOLOGIE::sommet: {
      pos_inversion (*dessin_pos, mark, nbf, nbs+nba);
      ancre_inversion (ancre0, mark, nbf, nbs+nba);
      ancre_inversion (ancre1, mark, nbf, nbs+nba);
      } break;
    case TOPOLOGIE::arete: {
      pos_inversion (*dessin_pos, mark, nbf+nbs, nba);
      ancre_inversion (ancre0, mark, nbf+nbs, nba);
      ancre_inversion (ancre1, mark, nbf+nbs, nba);
      } break;
    case TOPOLOGIE::quatrieme:
    case TOPOLOGIE::lambda:
      cerr << "ERROR: Never Reached\n";
      throw this;
  }
  dessin_topo->inversion (mark);
  delete dessin_liens; dessin_liens = new LIENS (topologie());
  liens_fixes     = false;
  moved ();
}

void
TOPOLOGIE::inversion (ulong mark)
{
  copy_if_multiref();
  const ulong nbf = faces().n();
  const ulong nbs = sommets().n();
  VALENCES* v = topo_faces;
  switch (typeramif(mark)) {
    case TOPOLOGIE::face:
      topo_faces -> swap(0, mark);
      break;
    case TOPOLOGIE::sommet:
      topo_sommets -> swap(0, mark-nbf);
      topo_faces   = topo_sommets;
      topo_sommets = topo_aretes;
      topo_aretes  = v;
      break;
    case TOPOLOGIE::arete:
      topo_aretes  -> swap(0, mark-nbf-nbs);
      topo_faces   = topo_aretes;
      topo_aretes  = topo_sommets;
      topo_sommets = v;
      break;
    case TOPOLOGIE::quatrieme:
    case TOPOLOGIE::lambda:
      cerr << "ERROR: Never Reached\n";
      throw this;
  }
}

void
VALENCES::swap (ulong i, ulong j)
{
  if ((i >= nombre) || (j >= nombre))
  { cerr << "VALENCES , swap out of range\n"; return; }
  ulong v = liste[i];
  liste[i] = liste[j];
  liste[j] = v;
}

void
DESSIN::precision (bool augmente)
{
  // empile ();
  cerr << "TODO Precision\n";
  // moved ();
}

// ----------------------------------------------------------------------
// class DESSIN::HISTORIQUE

DESSIN::HISTORIQUE::MEMOIRE::~MEMOIRE ()
{
  if (t) delete t;
  if (x) delete x;
}

DESSIN::HISTORIQUE::MEMOIRE::MEMOIRE ()
: t(NULL), x(NULL) { }

DESSIN::HISTORIQUE::~HISTORIQUE()
{
  delete[] hist_archives;
}

DESSIN::HISTORIQUE::HISTORIQUE(ulong d)
{
  hist_archives = new struct MEMOIRE [d];
  hist_maxdepth = d;
  hist_depth = 0;
}

void
DESSIN::HISTORIQUE::empile(const TOPOLOGIE& t, const VECTEUR& x,
  const ANCRE& a0, const ANCRE& a1)
{
  if (hist_depth < hist_maxdepth) hist_depth++;
  else { // On dcale
    delete hist_archives[0].t;
    delete hist_archives[0].x;
    delete hist_archives[0].a0;
    delete hist_archives[0].a1;
    for (ulong i = 0; i < hist_maxdepth - 1; i++) {
      hist_archives[i].t = hist_archives[i+1].t;
      hist_archives[i].x = hist_archives[i+1].x;
      hist_archives[i].a0 = hist_archives[i+1].a0;
      hist_archives[i].a1 = hist_archives[i+1].a1;
    }
  }
  hist_archives[hist_depth-1].t = new TOPOLOGIE (t);
  hist_archives[hist_depth-1].x = new VECTEUR   (x);
  hist_archives[hist_depth-1].a0 = new ANCRE (a0);
  hist_archives[hist_depth-1].a1 = new ANCRE (a1);
}

void
DESSIN::HISTORIQUE::depile(TOPOLOGIE*& t, VECTEUR*& x, ANCRE*& a0, ANCRE*& a1)
{
  if (!hist_depth) {
    cout << "Historique vide !\n";
    return;
  }
  delete t;
  delete x;
  delete a0;
  delete a1;
  hist_depth--;
  t = new TOPOLOGIE (*hist_archives[hist_depth].t);
  x = new VECTEUR   (*hist_archives[hist_depth].x);
  a0 = new ANCRE    (*hist_archives[hist_depth].a0);
  a1 = new ANCRE    (*hist_archives[hist_depth].a1);
  hist_archives[hist_depth].t = NULL;
  hist_archives[hist_depth].x = NULL;
  hist_archives[hist_depth].a0 = NULL;
  hist_archives[hist_depth].a1 = NULL;
}

// ----------------------------------------------------------------------
// class BELYI

BELYI::BELYI (const TOPOLOGIE& t, const VECTEUR& x)
{
  __RE_GARBAGE_BEGIN
  ulong offset = 0;
  R = computepol(t.faces(),      x, 1, offset); offset += t.faces().n();
  P = computepol(t.sommets(),    x, 0, offset); offset += t.sommets().n();
  Q = computepol(t.aretes(),     x, 0, offset); offset += t.aretes().n();
  S = computepol(t.quatriemes(), x, 0, offset);
  lambda = x[0];
#if defined(mplib_pari)
  ulong nbpts = R.degre()+P.degre()+Q.degre()+S.degre()+5;
  GEN *gptr [2*nbpts];
  {for (ulong i = 0; i <= R.degre(); i++)
    genptr (gptr, i, R[i]);}
  {for (ulong i = 0; i <= P.degre(); i++)
    genptr (gptr, i+R.degre()+1, P[i]);}
  {for (ulong i = 0; i <= Q.degre(); i++)
    genptr (gptr, i+R.degre()+P.degre()+2, Q[i]);}
  {for (ulong i = 0; i <= S.degre(); i++)
    genptr (gptr, i+R.degre()+P.degre()+Q.degre()+3, S[i]);}
  genptr (gptr, R.degre()+P.degre()+Q.degre()+S.degre()+4, lambda);
  gerepilemany(avma_garbage, gptr, 2*nbpts);
#endif
}

const BELYI&
DESSIN::belyi ()
{
  if (! dessin_belyi) dessin_belyi = new BELYI (topologie(), position());
  return *dessin_belyi;
}


// ----------------------------------------------------------------------
// class DESSIN : belyi

// Le systme (dessin de degr d)
// - si 3 valeurs de ramification : d + 2 inconnues
//   quation - P + Q + lambda * R
//   2 ancres
// - si 4 valeurs de ramification : 2 d + 2 inconnues
//   quation - P + Q + lambda * R
//   quation - P + S + C * lambda * R
//   2 ancres
// On calcule l'oppos de l'image et la drive

POLYNOME
computepol (const VALENCES& exponents, const VECTEUR& roots,
                    ulong offset, ulong offsetroot)
{
  if (!exponents.n()) return POLYNOME(CX(0));
  POLYNOME p (CX(1));
  ulong i; for (i = offset; i < exponents.n(); i++)
    p.mul_xan(roots[i+offsetroot], exponents[i]);
  return p;
}

CX
negvaleur (const ANCRE& a, const VECTEUR& x)
{
  switch (a.type) {
    case ANCRE::nondefini:
         cerr << "Erreur : ancre non dfinie\n";
         return CX(0);
         break;
    case ANCRE::point:
         return - x[a[0]];
         break;
    case ANCRE::somme:
    case ANCRE::moyenne: {
         CX ret (0);
         for (ulong j = 0; j < a.nombre(); j++) ret += x[a[j]];
         if (a.type == ANCRE::moyenne) ret /= (1. * a.nombre());
         return - ret; }
         break;
    case ANCRE::produit: {
         CX ret (1);
         for (ulong j = 0; j < a.nombre(); j++) ret *= x[a[j]];
         return - ret; }
         break;
  }
  cerr << "ERROR: Never Reached\n";
  return CX(0);
}

CX
derivee (const ANCRE& a, const VECTEUR& x, ulong i)
{
  switch (a.type) {
    case ANCRE::nondefini:
         cerr << "Erreur : ancre non dfinie\n";
         return CX(0);
         break;
    case ANCRE::point:
         if (a[0] == i) return CX(1);
         else           return CX(0);
         break;
    case ANCRE::somme:
    case ANCRE::moyenne:
         {for (ulong j = 0; j < a.nombre(); j++)
           if (i == a[j]) 
             return (a.type == ANCRE::moyenne) ? CX(1./a.nombre()) : CX(1);}
         return CX(0);
         break;
    case ANCRE::produit:
         {for (ulong j = 0; j < a.nombre(); j++)
           if (i == a[j]) {
           CX ret (1);
           for (ulong k = 0; k < a.nombre(); k++)
             if (k != i) ret *= x[k];
           return ret;
         }}
         return CX(0);
         break;
  }
  cerr << "ERROR: Never Reached\n";
  return CX(0);
}

VECTEUR
DESSIN::negimage (const BELYI& b, const VECTEUR& x) const
{
  // quation   - P + Q + lambda * R
  POLYNOME eq0 = b.P - b.Q - b.lambda * b.R;
  VECTEUR fx = eq0.vecteur(b.P.degre());
  if (topologie().nombrevaleurs() == 4) {
  // quation   - P + S + C * lambda * R
    POLYNOME eq1 = b.P - b.S - topologie().C() * b.lambda * b.R;
    for (ulong i = 0; i < b.P.degre(); i++)
      fx += eq1[i];
  }
  // 2 ancres
  fx += negvaleur(*ancre0, x);
  fx += negvaleur(*ancre1, x) + CX(1);
  return fx;
}

const VECTEUR&
DESSIN::negimage ()
{
  if (dessin_negimage) return *dessin_negimage;
  dessin_negimage = new VECTEUR( negimage (belyi(), position()) );
  return *dessin_negimage;
}

// Calcul de la drive
// Mthode "fast"     : dP/dp = (1/x) ( -v P + p dP/dp )
// Mthode "tangente" : calcul de (f(x+e)-f(x))/e

const MCARREE&
DESSIN::derivee ()
{
  if (dessin_derivee) return *dessin_derivee;
  __RE_GARBAGE_BEGIN
  ulong dim = position().length();
  dessin_derivee = new MCARREE (dim);
  belyi(); // Il faut que les polynmes aient t calculs
  {for (ulong i = 0; i < dim; i++) {
    VECTEUR colder = derivee_fast (i);
    for (ulong j = 0; j < dim; j++)
      (*dessin_derivee)(j,i) = colder[j];
  }}
#if defined(mplib_pari)
  GEN *gptr [2*dim*dim];
  {for (ulong i = 0; i < dim; i++)
    for (ulong j = 0; j < dim; j++)
      genptr (gptr, j+i*dim,  (*dessin_derivee)(j,i));}
  gerepilemany(avma_garbage, gptr, 2*dim*dim);
#endif
  return *dessin_derivee;
}

// calcul de dP/dp = (1/x) ( -v P + p dP/dp )
VECTEUR
DESSIN::derivee_fast (ulong i) const
{
  const BELYI& b = *dessin_belyi;
  ulong dim = dessin_derivee->dim();
  ulong deg = b.P.degre();
  if (topologie().nombrevaleurs() == 4) { assert (dim==2*deg+2); }
  else { assert (dim==deg+2); }
  if (i==0) {
  // drive par rapport  lambda
    VECTEUR xx(dim);
    // quation   - P + Q + lambda * R
    {for (ulong j=0; j<deg; j++)
        xx[j] = b.R[j];}
    if (topologie().nombrevaleurs() == 4)
      {for (ulong j=deg; j<2*deg; j++)
        xx[j] = b.R[j-deg] * topologie().C();}
    xx[dim-1] = 0;
    xx[dim-2] = 0;
    return xx;
  }
  // drive par rapport  la position d'un point
  POLYNOME *p; // polynme faisant intervenir cette variable
  switch (topologie().typeramif(i)) {
    case TOPOLOGIE::face:      p = new POLYNOME(b.R); (*p) *= b.lambda; break;
    case TOPOLOGIE::sommet:    p = new POLYNOME(b.P); (*p) *= CX (-1); break;
    case TOPOLOGIE::arete:     p = new POLYNOME(b.Q); break;
    case TOPOLOGIE::quatrieme: p = new POLYNOME(b.S); break;
    case TOPOLOGIE::lambda:    throw invalid; break;
  }
  VECTEUR xx(dim);
  // quation   - P + Q + lambda * R
  if (topologie().typeramif(i) == TOPOLOGIE::quatrieme)
    for (ulong j = deg; j; j--) xx[j-1] = CX(0);
  else {
    const POLYNOME& pol = *p;
    for (ulong j = deg; j; j--) {
      xx[j-1] = pol[j];
      xx[j-1] *= - static_cast<long>( topologie().valence(i) );
      if (j != deg) xx[j-1] += xx[j] * position()[i];
    }
  }
  if (topologie().nombrevaleurs() == 4)
  if (topologie().typeramif(i) == TOPOLOGIE::arete)
    for (ulong j = 2*deg; j>deg; j--) xx[j-1] = CX(0);
  else {
  // quation   - P + S + C * lambda * R
    if (topologie().typeramif(i) == TOPOLOGIE::face)
      (*p) *= topologie().C();
    const POLYNOME& pol = *p;
    for (ulong j = 2*deg; j>deg; j--) {
      xx[j-1] = pol[j-deg];
      xx[j-1] *= - static_cast<long>( topologie().valence(i) );
      if (j != 2*deg) xx[j-1] += xx[j] * position()[i];
    }
  }
  // 2 ancres
  xx[dim-1] = ::derivee(*ancre1, position(), i);
  xx[dim-2] = ::derivee(*ancre0, position(), i);
  delete p;
  return xx;
}

// calcul de (f(x+e)-f(x))/e
VECTEUR
DESSIN::derivee_tangente (ulong i) const
{
  VECTEUR* I;
  if (dessin_negimage) I = dessin_negimage;
  else if (dessin_belyi) I = new VECTEUR (negimage (*dessin_belyi, position()));
  else I = new VECTEUR (negimage (BELYI (topologie(), position()), position()));
  CX epsilon = 0.01;
  VECTEUR xe = position(); xe[i] += epsilon;
  VECTEUR fxe = negimage (BELYI(topologie(), xe), xe);
  fxe -= *I; fxe /= -epsilon;
  if (!dessin_negimage) delete I;
  return fxe;
}

static const DESSIN* currentDESSIN = NULL;
void set_currentDESSIN(const DESSIN* d) { currentDESSIN = d; }
VECTEUR
negimage_currentDESSIN (const VECTEUR& x)
{
  if (currentDESSIN == NULL) return VECTEUR(0);
  return currentDESSIN->negimage (BELYI(currentDESSIN->topologie(), x), x);
}
